
# THIS FILE IS GENERATED FROM CVXPY SETUP.PY
short_version = '1.9.1'
version = '1.9.1'
full_version = '1.9.1'
git_revision = '4489dd6'
commit_count = '268'
release = True
if not release:
    version = full_version
